<?php declare(strict_types=1);

namespace Infomaker\Everyware\Base\OpenContent;

use Infomaker\Everyware\Support\Date;
use Infomaker\Everyware\Support\Str;

/**
 * QueryBuilder
 *
 * @link    http://infomaker.se
 * @package Infomaker\Everyware\Base\OpenContent
 * @since   Infomaker\Everyware\Base\OpenContent\QueryBuilder 1.0.0
 */
class QueryBuilder
{

    /**
     * Contains the query to use against Open Content
     *
     * @var string
     */
    private $queryString;

    /**
     * Contains a string to be used as text-search and will therefor be escaped
     *
     * @var string
     */
    private $text;

    public function __construct(string $query, string $text = '')
    {
        $this->setQuery($query);
        $this->setText($text);
    }

    public function append(string $query, string $glue = 'AND'): self
    {
        $query = trim($query);

        if (Str::isEmpty($query)) {
            return $this;
        }

        if (Str::isEmpty($this->queryString)) {
            return $this->setQuery($query);
        }

        return $this->setQuery(Str::append($query, $this->getQuery(), " {$glue} "));
    }

    public function prepend(string $query, string $glue = 'AND' ): self
    {
        $query = trim($query);

        if (Str::isEmpty($query)) {
            return $this;
        }

        if (Str::isEmpty($this->queryString)) {
            return $this->setQuery($query);
        }

        return $this->setQuery(Str::prepend($query, $this->getQuery(), " {$glue} "));
    }

    public function buildQueryString(): string
    {
        $text = $this->getSanitizedText();
        $query = $this->getQuery();

        return Str::notEmpty($text) ? Str::prepend("({$text}) AND", $query, ' ') : $query;
    }

    public function andIf(string $query): self
    {
        return $this->append($query, 'AND');
    }

    public function andIfNot($query): self
    {
        return $this->append($query, 'AND NOT');
    }

    public function orIf($query): self
    {
        return $this->append($query, 'OR');
    }

    public function orIfNot($query): self
    {
        return $this->append($query, 'OR NOT');
    }

    public function isEmpty(): bool
    {
        return Str::isEmpty($this->buildQueryString());
    }

    /**
     * Helper function for appending conditional query string
     *
     * @param string       $property The name of the property to match with
     * @param string|array $values   One or a list of values to look for
     * @param string       $glue
     *
     * @return self
     */
    public function andIfProperty(string $property, $values, string $glue = 'OR'): self
    {
        return $this->andIf(static::propertySearch($property, $values, $glue));
    }


    /**
     * Helper function for appending conditional query string
     *
     * @param string       $property The name of the property to match with
     * @param string|array $values   One or a list of values to look for
     * @param string       $glue
     *
     * @return self
     */
    public function andIfNotProperty(string $property, $values, string $glue = 'OR'): self
    {
        return $this->andIfNot(static::propertySearch($property, $values, $glue));
    }

    /**
     * Helper function for appending conditional query string
     *
     * @param string       $property The name of the property to match with
     * @param string|array $values   One or a list of values to look for
     * @param string       $glue
     *
     * @return self
     */
    public function orIfProperty($property, $values, $glue = 'OR'): self
    {
        return $this->orIf(static::propertySearch($property, $values, $glue));
    }

    /**
     * Helper function for appending conditional query string
     *
     * @param string       $property The name of the property to match with
     * @param string|array $values   One or a list of values to look for
     * @param string       $glue
     *
     * @return self
     */
    public function orIfNotProperty($property, $values, $glue = 'OR'): self
    {
        return $this->orIfNot(static::propertySearch($property, $values, $glue));
    }

    public function getQuery(): string
    {
        return $this->queryString;
    }

    public function getText(): string
    {
        return $this->text;
    }

    public function getSanitizedText(): string
    {
        $text = stripcslashes($this->getText());

        return Str::notEmpty($text) ? '"' . addslashes($text) . '"' : $text;
    }

    public function setText(string $text): self
    {
        $this->text = trim($text);

        return $this;
    }

    public function setQuery(string $query): self
    {
        $this->queryString = trim($query);

        return $this;
    }
    public function __toString()
    {
        return $this->buildQueryString();
    }

    public static function createProperty(string $name, $value): string
    {
        if (\is_bool($value)) {
            $value = $value ? 'true' : 'false';
        }

        $value = Str::contains($value, ' ') ? '"' . $value . '"' : $value;

        return "{$name}:{$value}";
    }

    /**
     * Create a date readable by Open Content
     *
     * @param string $date
     *
     * @return string
     */
    public static function createOcDate($date = ''): string
    {
        return Date::parse($date)->toOcString();
    }

    /**
     * Create a Solr-query block indicating a date-span
     *
     * @param string $property
     * @param string $from
     * @param string $to
     *
     * @return string
     */
    public static function dateSpanSearch($property, $from = '', $to = ''): string
    {
        $spanStart = Str::notEmpty($from) ? static::createOcDate($from) : '*';
        $spanStop = Str::notEmpty($to) ? static::createOcDate($to) : 'NOW';

        return "{$property}:[{$spanStart} TO {$spanStop}]";
    }

    /**
     * Creates a Solr-query of one or more properties with the option to change the separator
     *
     * @param string            $property The name of the property to match with
     * @param string|bool|array $values   One or a list of values to look for
     * @param string            $glue
     *
     * @return string
     */
    public static function propertySearch($property, $values, $glue = 'OR'): string
    {
        $values = array_unique(array_filter((array)$values));

        if ( ! empty($values)) {
            $glue = trim($glue); // Make sure that there is no space around the glue

            $block = implode(" {$glue} ", array_map(function ($value) use ($property) {
                return static::createProperty($property, $value);
            }, $values));

            return \count($values) > 1 ? "({$block})" : $block;
        }

        return '';
    }

    /**
     * Start a query by property search
     *
     * @param string       $property The name of the property to match with
     * @param string|array $values   One or a list of values to look for
     * @param string       $glue
     *
     * @return self
     */
    public static function where($property, $values, $glue = 'OR'): self
    {
        return static::query(static::propertySearch($property, $values, $glue));
    }

    /**
     * Start a query
     *
     * @param string $query
     *
     * @return self
     */
    public static function query($query): self
    {
        return new static($query);
    }

    /**
     * Start a query by adding text
     *
     * @param string $text
     *
     * @return self
     */
    public static function text($text): self
    {
        return new static('', $text);
    }
}
